class Q3RconLibError(Exception):
    """Base Q3RconLib error class"""


class Q3RconLibLoginError(Q3RconLibError):
    """Exception raised when a bad rcon login attempt is made"""

    err_msg = {
        1: "Invalid rcon password.",
        2: "Unable to contact the server, perhaps it's down?",
        3: "No rcon password set on server or password is shorter than 8 characters.",
    }

    def __init__(self, code):
        self.code = code
        super().__init__(self.err_msg[self.code])



from .packet import Packet


class Response(Packet):
    @property
    def header(self) -> bytes:
        return Packet.MAGIC + "print\n".encode()

    def decode(self, resp: bytes) -> str:
        return resp.removeprefix(self.header).decode()



# mypy: disable-error-code="attr-defined"

import logging
import socket

from .packet import Request, Response

logger = logging.getLogger(__name__)


class Base:
    def __init__(self, **kwargs):
        self.logger = logger.getChild(self.__class__.__name__)
        defaultkwargs = {
            "host": "localhost",
            "port": 28960,
            "password": "",
            "default_timeout": 0.02,
            "timeouts": {},
        }
        kwargs = defaultkwargs | kwargs
        for attr, val in kwargs.items():
            setattr(self, attr, val)

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._request = Request(self.password)
        self._response = Response()

    def send(self, cmd: str) -> str:
        """send a rcon command to the server

        Args:
            cmd (str): command string that takes the form "<cmd> <args>"

        Returns:
            str: the entire command response (all fragments) as one string
        """
        timeout: float = self.timeouts.get(
            cmd.split(" ", maxsplit=1)[0], self.default_timeout
        )
        self._sock.settimeout(timeout)

        self._sock.sendto(
            self._request.encode(cmd),
            (socket.gethostbyname(self.host), self.port),
        )
        self.logger.debug(f"sending: {cmd}")

        fragments: list[str] = []
        try:
            while resp := self._sock.recv(2048):
                if len(resp) > len(self._response.header):
                    if resp.startswith(self._response.header):
                        fragments.append(self._response.decode(resp))
        except TimeoutError:
            self.logger.debug("finished collecting response fragments")

        return "".join(fragments)

    def close(self):
        self._sock.close()



import functools
import time

from .error import Q3RconLibLoginError


def timeout(func):
    """Attempts a rcon login until time elapsed is greater than {Q3Rcon}._login_timeout"""

    @functools.wraps(func)
    def wrapper(*args):
        rcon, *_ = args

        err = None
        start = time.time()
        while time.time() < start + rcon._login_timeout:
            try:
                resp = func(*args)
                if not resp:
                    raise TimeoutError("timeout attempting to login")
                elif resp in (
                    "Bad rcon",
                    "Bad rconpassword.",
                    "Invalid password.",
                ):
                    raise Q3RconLibLoginError(code=1)
                elif (
                    resp == "No rcon password set on server or "
                    "password is shorter than 8 characters.\n"
                ):
                    raise Q3RconLibLoginError(code=3)

                err = None
                break
            except (TimeoutError, ConnectionRefusedError, ConnectionResetError) as e:
                err = e
                rcon.logger.error(f"{type(e).__name__}: {e}... retrying login attempt")
                continue

        if err:
            raise Q3RconLibLoginError(code=2)

    return wrapper



import logging
import socket
from typing import Self

from . import util
from .base import Base
from .error import Q3RconLibLoginError

logger = logging.getLogger(__name__)


class Q3Rcon:
    def __init__(self, **kwargs):
        self.logger = logger.getChild(self.__class__.__name__)
        self._login_timeout = kwargs.pop("login_timeout", 2)

        self._base = Base(**kwargs)

    def __enter__(self) -> Self:
        try:
            self._login()
        except socket.gaierror:
            self.logger.error(f"unable to resolve {self._base.host}")
            raise
        except Q3RconLibLoginError as e:
            self.logger.error(f"{type(e).__name__}: {e}")
            raise
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._base.close()

    def __repr__(self) -> str:
        return type(
            self
        ).__name__ + "(hostname='{host}', port={port}, password='{password}')".format(
            **self._base.__dict__
        )

    def __str__(self) -> str:
        return "Rcon connection to {host}:{port} with password {password}".format(
            **self._base.__dict__
        )

    @property
    def host(self) -> str:
        return self._base.host

    @property
    def port(self) -> int:
        return self._base.port

    @util.timeout
    def _login(self):
        return self.send("login")

    def send(self, cmd: str) -> str:
        return self._base.send(cmd)


def request_q3rcon_obj(**kwargs) -> Q3Rcon:
    """Interface entry point, use it to request a Q3Rcon object.

    Returns:
        Q3Rcon: A class implementing the Q3 Rcon protocol
    """
    cls = Q3Rcon
    return cls(**kwargs)



from .packet import Packet


class Request(Packet):
    def __init__(self, password: str):
        self.password = password

    @property
    def header(self) -> bytes:
        return Packet.MAGIC + "rcon".encode()

    def encode(self, cmd) -> bytes:
        return self.header + f" {self.password} {cmd}".encode()
